# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2011 Nicira Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# @author: Somik Behera, Nicira Networks, Inc.
# @author: Brad Hall, Nicira Networks, Inc.
# @author: Dan Wendlandt, Nicira Networks, Inc.

import re
import time

from eventlet import db_pool
from eventlet import greenthread
try:
    import MySQLdb
except ImportError:
    MySQLdb = None
from oslo.config import cfg
import sqlalchemy as sql
from sqlalchemy import create_engine
from sqlalchemy.exc import DisconnectionError
from sqlalchemy.exc import OperationalError
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.interfaces import PoolListener
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import NullPool
from sqlalchemy.pool import StaticPool

from quantum.db import model_base
from quantum.openstack.common import log as logging
from quantum.openstack.common.exception import DBError

LOG = logging.getLogger(__name__)
SQL_CONNECTION_DEFAULT = 'sqlite://'


database_opts = [
    cfg.StrOpt('sql_connection',
               help=_('The SQLAlchemy connection string used to connect to '
                      'the database'),
               secret=True),
    cfg.IntOpt('sql_max_retries', default=2,
               help=_('Database reconnection reconnect times')),
    cfg.IntOpt('reconnect_interval', default=1,
               help=_('Database reconnection interval in seconds')),
    cfg.BoolOpt('sql_inc_reconnect_interval',
                default=True,
                help='Whether to increase interval between sql connection'
                     'retries, up to sql_max_reconnect_interval'),
    cfg.IntOpt('sql_max_reconnect_interval',
               default=10,
               help='max seconds between sql connection retries, if '
               'sql_inc_retry_interval is enabled'),
    cfg.IntOpt('sql_min_pool_size',
               default=1,
               help=_("Minimum number of SQL connections to keep open in a "
                      "pool")),
    cfg.IntOpt('sql_max_pool_size',
               default=5,
               help=_("Maximum number of SQL connections to keep open in a "
                      "pool")),
    cfg.IntOpt('sql_idle_timeout',
               default=30,
               help=_("Timeout in seconds before idle sql connections are "
                      "reaped")),
    cfg.BoolOpt('sql_dbpool_enable',
                default=False,
                help=_("Enable the use of eventlet's db_pool for MySQL")),
    cfg.IntOpt('sqlalchemy_pool_size',
               default=None,
               help=_("Maximum number of SQL connections to keep open in a "
                      "QueuePool in SQLAlchemy")),
    cfg.IntOpt('sqlalchemy_max_overflow',
               default=None,
               help=_("If set, use this value for max_overflow with "
                      "sqlalchemy")),
    cfg.IntOpt('sqlalchemy_pool_timeout',
               default=None,
               help=_("If set, use this value for pool_timeout with "
                      "sqlalchemy")),
]

cfg.CONF.register_opts(database_opts, "DATABASE")

_ENGINE = None
_MAKER = None
BASE = model_base.BASEV2


class MySQLPingListener(object):

    """
    Ensures that MySQL connections checked out of the
    pool are alive.

    Borrowed from:
    http://groups.google.com/group/sqlalchemy/msg/a4ce563d802c929f
    """

    def checkout(self, dbapi_con, con_record, con_proxy):
        try:
            dbapi_con.cursor().execute('select 1')
        except dbapi_con.OperationalError, ex:
            if ex.args[0] in (2006, 2013, 2014, 2045, 2055):
                LOG.warn(_('Got mysql server has gone away: %s'), ex)
                raise DisconnectionError(_("Database server went away"))
            else:
                raise


class SqliteForeignKeysListener(PoolListener):
    """
    Ensures that the foreign key constraints are enforced in SQLite.

    The foreign key constraints are disabled by default in SQLite,
    so the foreign key constraints will be enabled here for every
    database connection
    """
    def connect(self, dbapi_con, con_record):
        dbapi_con.execute('pragma foreign_keys=ON')


def configure_db():
    """
    Establish the database, create an engine if needed, and
    register the models.
    """
    global _ENGINE
    if not _ENGINE:
        sql_connection = cfg.CONF.DATABASE.sql_connection
        if not sql_connection:
            LOG.warn(_("Option 'sql_connection' not specified "
                       "in any config file - using default "
                       "value '%s'" % SQL_CONNECTION_DEFAULT))
            sql_connection = SQL_CONNECTION_DEFAULT
        connection_dict = sql.engine.url.make_url(sql_connection)
        engine_args = {
            'pool_recycle': 3600,
            'echo': False,
            'convert_unicode': True,
        }

        if cfg.CONF.DATABASE.sqlalchemy_pool_size is not None:
            pool_size = cfg.CONF.DATABASE.sqlalchemy_pool_size
            engine_args['pool_size'] = pool_size
        if cfg.CONF.DATABASE.sqlalchemy_max_overflow is not None:
            max_overflow = cfg.CONF.DATABASE.sqlalchemy_max_overflow
            engine_args['max_overflow'] = max_overflow
        if cfg.CONF.DATABASE.sqlalchemy_pool_timeout is not None:
            pool_timeout = cfg.CONF.DATABASE.sqlalchemy_pool_timeout
            engine_args['pool_timeout'] = pool_timeout

        if 'mysql' in connection_dict.drivername:
            engine_args['listeners'] = [MySQLPingListener()]
            if (MySQLdb is not None and
                cfg.CONF.DATABASE.sql_dbpool_enable):
                pool_args = {
                    'db': connection_dict.database,
                    'passwd': connection_dict.password or '',
                    'host': connection_dict.host,
                    'user': connection_dict.username,
                    'min_size': cfg.CONF.DATABASE.sql_min_pool_size,
                    'max_size': cfg.CONF.DATABASE.sql_max_pool_size,
                    'max_idle': cfg.CONF.DATABASE.sql_idle_timeout
                }
                pool = db_pool.ConnectionPool(MySQLdb, **pool_args)

                def creator():
                    conn = pool.create()
                    # NOTE(belliott) eventlet >= 0.10 returns a tuple
                    if isinstance(conn, tuple):
                        _1, _2, conn = conn
                    return conn

                engine_args['creator'] = creator
            if (MySQLdb is None and cfg.CONF.DATABASE.sql_dbpool_enable):
                LOG.warn(_("Eventlet connection pooling will not work without "
                           "python-mysqldb!"))
        if 'sqlite' in connection_dict.drivername:
            engine_args['listeners'] = [SqliteForeignKeysListener()]
            if sql_connection == "sqlite://":
                engine_args["connect_args"] = {'check_same_thread': False}

        _ENGINE = create_engine(sql_connection, **engine_args)

        sql.event.listen(_ENGINE, 'checkin', greenthread_yield)

        if not register_models():
            if cfg.CONF.DATABASE.reconnect_interval:
                remaining = cfg.CONF.DATABASE.sql_max_retries
                reconnect_interval = cfg.CONF.DATABASE.reconnect_interval
                retry_registration(remaining, reconnect_interval)


def clear_db(base=BASE):
    global _ENGINE, _MAKER
    assert _ENGINE

    unregister_models(base)
    if _MAKER:
        _MAKER.close_all()
        _MAKER = None
    _ENGINE.dispose()
    _ENGINE = None


def is_db_connection_error(args):
    """Return True if error in connecting to db."""
    # NOTE(adam_g): This is currently MySQL specific and needs to be extended
    #               to support Postgres and others.
    conn_err_codes = ('2002', '2003', '2006', '2013', '1213')
    for err_code in conn_err_codes:
        if str(args).find(err_code) != -1:
            return True
    return False


def wrap_db_error(f):
    """Function wrapper to capture DB errors

    If an exception is thrown by the wrapped function,
    determine if it represents a database connection error.
    If so, retry the wrapped function, and repeat until it succeeds
    or we reach a configurable maximum number of retries.
    If it is not a connection error, or we exceeded the retry limit,
    raise a DBError.

    """

    def _wrap_db_error(*args, **kwargs):
        sql_connection = cfg.CONF.DATABASE.sql_connection
        if not sql_connection:
            LOG.warn(_("Option 'sql_connection' not specified "
                       "in any config file - using default "
                       "value '%s'" % SQL_CONNECTION_DEFAULT))
            sql_connection = SQL_CONNECTION_DEFAULT
        connection_dict = sql.engine.url.make_url(sql_connection)
        if 'mysql' in connection_dict.drivername:
            import MySQLdb
            operational_errors = (MySQLdb.OperationalError, OperationalError)
        else:
            operational_errors = (OperationalError,)
        next_interval = cfg.CONF.DATABASE.reconnect_interval
        remaining = cfg.CONF.DATABASE.sql_max_retries
        if remaining == -1:
            remaining = 'infinite'
        while True:
            try:
                return f(*args, **kwargs)
            except UnicodeEncodeError:
                raise InvalidUnicodeParameter()
            except operational_errors as e:
                if is_db_connection_error(e.args[0]):
                    if remaining == 0:
                        LOG.exception(_('DB exceeded retry limit.'))
                        raise DBError(e)
                    if remaining != 'infinite':
                        remaining -= 1
                    LOG.exception(_('DB connection error, '
                                    'retrying in %i seconds.') % next_interval)
                    time.sleep(next_interval)
                    if cfg.CONF.DATABASE.sql_inc_reconnect_interval:
                        next_interval = min(
                            next_interval * 2,
                            cfg.CONF.DATABASE.sql_max_reconnect_interval)
                else:
                    LOG.exception(_('DB exception wrapped.'))
                    raise DBError(e)
            except InvalidRequestError as e:
                exc_re = re.search('(.*)Original\s+exception\s+was:'
                                   '\s+\((.*)\)\s+\((.*)\)', e.args[0])
                if exc_re:
                    invalidreqerror_msg = exc_re.group(1)
                    original_exception_type = exc_re.group(2)
                    original_exception_value = exc_re.group(3)
                    if re.search('rollback', invalidreqerror_msg):
                        get_session().rollback()
                    if is_db_connection_error(original_exception_value):
                        if remaining == 0:
                            LOG.exception(_('DB exceeded retry limit.'))
                            raise DBError(e)
                        if remaining != 'infinite':
                            remaining -= 1
                        LOG.exception(_('DB connection error, '
                                        'retrying in %i seconds.')
                                      % next_interval)
                        time.sleep(next_interval)
                        if cfg.CONF.DATABASE.sql_inc_reconnect_interval:
                            next_interval = min(
                                next_interval * 2,
                                cfg.CONF.DATABASE.sql_max_reconnect_interval)
                    else:
                        LOG.exception(_('DB exception wrapped.'))
                        raise DBError(e)
                else:
                    LOG.exception(_('DB exception wrapped.'))
                    raise DBError(e)
            except Exception, e:
                LOG.exception(_('DB exception wrapped.'))
                raise DBError(e)

    _wrap_db_error.func_name = f.func_name
    return _wrap_db_error


def get_session(autocommit=True, expire_on_commit=False):
    """Helper method to grab session"""
    class OurQuery(sql.orm.query.Query):
        pass

    global _MAKER, _ENGINE
    if not _MAKER:
        assert _ENGINE
        query = OurQuery
        query.all = wrap_db_error(query.all)
        query.first = wrap_db_error(query.first)
        query.update = wrap_db_error(query.update)
        query.delete = wrap_db_error(query.delete)
        _MAKER = sessionmaker(bind=_ENGINE,
                              autocommit=autocommit,
                              expire_on_commit=expire_on_commit,
                              query_cls=query)
    session = _MAKER()
    session.begin = wrap_db_error(session.begin)
    session.execute = wrap_db_error(session.execute)
    session.flush = wrap_db_error(session.flush)
    session.query = wrap_db_error(session.query)
    session.delete = wrap_db_error(session.delete)
    session.commit = wrap_db_error(session.commit)
    session.rollback = wrap_db_error(session.rollback)
    session.expire_all()
    return session


def retry_registration(remaining, reconnect_interval, base=BASE):
    if remaining == -1:
        remaining = 'infinite'
    while True:
        if remaining != 'infinite':
            if remaining == 0:
                LOG.error(_("Database connection lost, exit..."))
                break
            remaining -= 1
        LOG.info(_("Unable to connect to database, %(remaining)s attempts "
                   "left. Retrying in %(reconnect_interval)s seconds"),
                 locals())
        time.sleep(reconnect_interval)
        if register_models(base):
            break


def register_models(base=BASE):
    """Register Models and create properties"""
    global _ENGINE
    assert _ENGINE
    try:
        base.metadata.create_all(_ENGINE)
    except sql.exc.OperationalError as e:
        LOG.info(_("Database registration exception: %s"), e)
        return False
    return True


def unregister_models(base=BASE):
    """Unregister Models, useful clearing out data before testing"""
    global _ENGINE
    assert _ENGINE
    base.metadata.drop_all(_ENGINE)


def greenthread_yield(dbapi_con, con_record):
    """
    Ensure other greenthreads get a chance to execute by forcing a context
    switch. With common database backends (eg MySQLdb and sqlite), there is
    no implicit yield caused by network I/O since they are implemented by
    C libraries that eventlet cannot monkey patch.
    """
    greenthread.sleep(0)
